/*
 * video.c: A plugin for the Video Disk Recorder
 *
 * See the README file for copyright information and how to reach the author.
 *
 * $Id: video.c,v 1.6 2004/05/22 07:52:39 stefan Exp $
 */

#include <sys/mman.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <vdr/plugin.h>
#include "video.h"
#include "utils.h"


cVideoOut::~cVideoOut()
{
  dsyslog("[video]: Good bye");
}

#if VDRVERSNUM >= 10307

void cVideoOut::OpenOSD(int X, int Y)
{
  OSDxOfs = X;
  OSDyOfs = Y;
  OSDpresent=true;
}

void cVideoOut::CloseOSD()
{
  OSDpresent=false;
}

void cVideoOut::Draw(cBitmap *Bitmap, unsigned char *osd_buf, int linelen)
{
    int           depth = (Bpp + 7) / 8;
    int           a, r, g, b;
    tColor        c;
    tIndex        *buf;
    const tIndex  *adr;

  for (int y = 0; y < Bitmap->Height(); y++)
  {
    buf = (tIndex *) osd_buf +
            linelen * ( OSDyOfs + y ) +
            OSDxOfs * depth;

    for (int x = 0; x < Bitmap->Width(); x++)
    {
      adr = Bitmap->Data(x, y);
      c = Bitmap->Color(*adr);
      a = (c >> 24) & 255; //Alpha
      r = (c >> 16) & 255; //Red
      g = (c >> 8) & 255;  //Green
      b = c & 255;         //Blue
      switch (depth) {
        case 4:
          if (a == 255 && r == 0 && g == 0 && b == 0)
          {
            buf[0] = 1; buf[1] = 1; buf[2] = 1;
          } else {
            buf[0] = b;
            buf[1] = g;
            buf[2] = r;
            buf[3] = a;
          }
          buf += 4;
          break;
        case 3:
          if (a == 255 && r == 0 && g == 0 && b == 0)
          {
            buf[0] = 1; buf[1] = 1; buf[2] = 1;
          } else {
            buf[0] = b;
            buf[1] = g;
            buf[2] = r;
          }
          buf += 3;
          break;
        case 2:
          if (a == 255 && r == 0 && g == 0 && b == 0)
          {
              buf[0] = 0x21; buf[1] = 0x08;
          } else {
            buf[0] = ((b >> 3)& 0x1F) | ((g & 0x1C) << 3);
            buf[1] = (r & 0xF8) | (g >> 5);
          }
          buf += 2;
          break;
        default:
            dsyslog("[video] OSD: unsupported depth %d exiting",depth);
            exit(1);
          break;
      }
    }
  }
}

#else

bool cVideoOut::OpenWindow(cWindow *Window) {
    layer[Window->Handle()]= new cWindowLayer(Window->X0()+OSDxOfs, Window->Y0()+OSDyOfs,
	Window->Width(), Window->Height(), Bpp, Xres, Yres);
    return true;
}

void cVideoOut::OpenOSD(int X, int Y)
{
  // initialize Layers.
  OSDxOfs = (Xres - 720) / 2 + X;
  OSDyOfs = (Yres - 576) / 2 + Y;
  osdMutex.Lock();
  for (int i = 0; i < MAXNUMWINDOWS; i++)
  {
    layer[i]=0;
  }
  OSDpresent=true;
  osdMutex.Unlock();
}

void cVideoOut::CloseOSD()
{
  osdMutex.Lock();
  OSDpresent=false;
  for (int i = 0; i < MAXNUMWINDOWS; i++)
  {
    if (layer[i])
    {
      delete(layer[i]);
      layer[i]=0;
    }
  }
  osdMutex.Unlock();
}

void cVideoOut::CommitWindow(cWindow *Window) {
    layer[Window->Handle()]->Render(Window);
    Refresh();
}

void cVideoOut::ShowWindow(cWindow *Window) {
    layer[Window->Handle()]->visible=true;
    layer[Window->Handle()]->Render(Window);
    Refresh();
}
void cVideoOut::HideWindow(cWindow *Window, bool Hide) {
    layer[Window->Handle()]->visible= ! Hide ;
    Refresh();
}

void cVideoOut::MoveWindow(cWindow *Window, int x, int y) {
    layer[Window->Handle()]->Move(x,y);
    layer[Window->Handle()]->Render(Window);
    Refresh();
}

void cVideoOut::CloseWindow(cWindow *Window) {
    delete (layer[Window->Handle()]);
    Refresh();
}


// --- cWindowLayer --------------------------------------------------
cWindowLayer::cWindowLayer(int X, int Y, int W, int H, int Bpp, int Xres, int Yres) {
    left=X;
    top=Y;
    width=W;
    height=H;
    bpp=Bpp;
    xres=Xres;
    yres=Yres;
    visible=false;
    imagedata=(unsigned char *)malloc(W*H*4); // RGBA Screen memory
    printf("[video] Creating WindowLayer at %d x %d, (%d x %d)\n",X,Y,W,H);
}

void cWindowLayer::Region (int *x, int *y, int *w, int *h) {
   *x = left;
   *y = top;
   *w = width;
   *h = height;
}


cWindowLayer::~cWindowLayer() {
    free(imagedata);
}


void cWindowLayer::Render(cWindow *Window) {
    unsigned char * buf;
    buf=imagedata;
    for (int yp = 0; yp < height; yp++) {
	for (int ix = 0; ix < width; ix++) {
	    eDvbColor c = Window->GetColor(*Window->Data(ix,yp));
	    buf[0]=c & 255; //Red
	    buf[1]=(c >> 8) & 255; //Green
	    buf[2]=(c >> 16) & 255; //Blue
	    buf[3]=(c >> 24) & 255; //Alpha*/
	    buf+=4;
	}
    }
}

void cWindowLayer::Move(int x, int y) {
    left=x;
    top=y;
}

void cWindowLayer::Draw(unsigned char * buf, int linelen, unsigned char * keymap) {
    unsigned char * im;
    im = imagedata;
    int depth = (bpp + 7) / 8;
    int dx = linelen - width * depth;
    buf += top * linelen + left * depth; // upper left corner
    for (int y = top; y < top+height; y++) {
	for (int x = left; x < left+width; x++) {
	   if ( (im[3] != 0)
		&& (x >= 0) && (x < xres)
		&& (y >= 0) && (y < yres))  { // Alpha != 0 and in the screen

		//if (keymap) keymap[(x+y*linelen / depth) / 8] |= (1 << (x % 8));
		switch (depth) {
		    case 4:
          if (im[3] == 255 && im[0] == 0 && im[1] == 0 && im[2] == 0) {
            *buf++ = 1; *buf++ = 1; *buf++ = 1; *buf++ = 255;
          } else {
            *(buf++)=im[2];
            *(buf++)=im[1];
            *(buf++)=im[0];
            *(buf++)=im[3];
          }
			//buf++;
			break;
		    case 3:
          if (im[3] == 255 && im[0] == 0 && im[1] == 0 && im[2] == 0) {
            *buf++ = 1; *buf++ = 1; *buf++ = 1;
          } else {
            *(buf++)=im[2];
            *(buf++)=im[1];
            *(buf++)=im[0];
          }
			break;
		    case 2: // 565 RGB
          if (im[3] == 255 && im[0] == 0 && im[1] == 0 && im[2] == 0) {
            *buf++ = 0x21; *buf++ = 0x08;
          } else {
            *(buf++)= ((im[2] >> 3)& 0x1F) | ((im[1] & 0x1C) << 3);
            *(buf++)= (im[0] & 0xF8) | (im[1] >> 5);
          }
			break;
		    default:
			dsyslog("[video] Unsupported depth %d exiting",depth);
			exit(1);
    		}

	    } else  {
	        buf += depth; // skip this pixel
	    }
	    im +=4;
        }
        buf += dx;
    }
    return;
}

#endif
